# -*- coding: utf-8 -*-
"""
Created on Wed May 24 11:19:36 2023

@author: kruis
"""

import numpy as np
import os
import pandas as pd
from scipy.signal import savgol_filter, argrelextrema
from sklearn.cluster import DBSCAN
import math
import rasterio.features
import geopandas as gpd
from shapely.geometry import LineString

"""
This function reconstructs a 2D signal from the different fourier wavelengths  
    in the Wavelet spectrum.
    
Arguments:
- rec3d: a 3D numpy array representing a reconstructed signal (dimensions: xs, 
    ds, wavelengths).
- Zavg: a 2D numpy array representing the elevation signal detrended by the 
    average bed level (dimensions: xs, ds).
    
Returns:
- Zrec: a 2D numpy array representing the reconstructed wavelet signal 
    (dimensions: xs, ds).
"""
# Reconstruct the signal to correspond with the original signal
def reconstruct(rec3d, Zavg):
    # Sum all frequencies of the signal
    Zrec = rec3d.sum(axis=2)
    # Ensure that the non measured points are removed from the reconstruction
    Zrec[np.isnan(Zavg)] = np.nan
    # Normalize the reconstruction to [0, 1] range
    Zrec = 1.0 - (Zrec - np.nanmin(Zrec)) / (np.nanmax(Zrec) - np.nanmin(Zrec))
    # Scale reconstruction to the range of Zavg
    Zrec = Zrec * (np.nanmax(Zavg)-np.nanmin(Zavg))
    # Shift reconstruction to match the mean of Zavg
    Zrec = Zrec - (np.nanmean(Zrec)-np.nanmean(Zavg))
    return Zrec

"""
This function applies a Savitzky-Golay filter to smoothen a numpy array.

Arguments:
- arr: a numpy array to be smoothened.

Returns:
- arrS: a smoothened numpy array.
"""
def smoothen(arr):
    arrS = np.nan_to_num(arr)
    arrS = savgol_filter(arrS,41,3,mode='mirror',axis = 0)
    arrS = np.where(np.isnan(arr), np.nan, arrS) 
    return arrS


"""
This function assigns cluster labels (duneNumbers) to 2D data points using the 
    DBSCAN algorithm.

Arguments:
- df: a pandas dataframe containing 2D data points in columns 'ds' and 'xs'.
- eps: the maximum distance between two points to be considered as in the same 
    cluster.
- min_samples: the minimum number of points required to form a cluster.

Returns:
- df: the updated pandas dataframe with an additional column 'number' 
    containing the assigned cluster labels (duneNumbers).
"""
# Define a function to assign duneNumber based on clustering
def assign_dune_numbers(df, eps=5, min_samples=5):
    # Convert dataframe to a numpy array of (x, y) coordinates
    coords = df[['ds', 'xs']].to_numpy()
    # Cluster the points using DBSCAN
    db = DBSCAN(eps=eps, min_samples=min_samples)
    labels = db.fit_predict(coords)
    # Assign the cluster labels as duneNumbers in the dataframe
    df['number'] = labels
    # Return the updated dataframe
    return df


"""
This function takes the original bathymetry, and fits a window around a center 
    point (xs, ds). The window dimensions are determined based on the range 
    size (range_size_xs, range_size_ds). It returns a  2D array of size 
    (2range_size_xs+1, 2range_size_ds+1) representing the window. It is a slice 
    of the original array, padded with NaN values where needed.
    
Arguments:
- orig_array: 2D array of shape (n_rows, n_cols), the original array.
- xs: integer, the row index of the center point.
- ds: integer, the column index of the center point.
- range_size_xs: integer, the range size in the row direction.
- range_size_ds: integer, the range size in the column direction.

Returns:
- window: 2D array of shape (2range_size_xs+1, 2range_size_ds+1), a slice of 
    the original array centered at the specified point.
"""
def getWindow(orig_array, xs, ds, range_size_xs, range_size_ds):
    # Define maximum values for xs and ds
    XS = orig_array.shape[0]
    DS = orig_array.shape[1]
    
    # Define the start and end indices for the rows and columns to slice
    row_start = max(0, xs - range_size_xs)
    row_end = min(XS, xs + range_size_xs + 1)  # Add 1 to include the last row
    col_start = max(0, ds - range_size_ds)
    col_end = min(DS, ds + range_size_ds + 1)  # Add 1 to include the last column
    
    # Slice the original array to extract the range of cells
    range_array = orig_array[row_start:row_end, col_start:col_end]
    
    # Calculate the padding required to make the new array of size (2*range_size+1, 2*range_size+1)
    pad_rows_before = max(0, range_size_xs - xs)
    pad_rows_after = max(0, range_size_xs - (XS - xs - 1))
    pad_cols_before = max(0, range_size_ds - ds)
    pad_cols_after = max(0, range_size_ds - (DS - ds - 1))
    
    # Pad the range array with NaN values to create the new array of size (2*range_size+1, 2*range_size+1)
    window = np.pad(range_array, ((pad_rows_before, pad_rows_after), (pad_cols_before, pad_cols_after)), mode='constant', constant_values=np.nan)
    return window


"""
This function converts two columns of a dataframe representing stream coordinates 
    and converts them to XY coordinates based on a predefined coversion dataframe.
    
Arguments:
- df: pd.DataFrame containing the coordinates that should be converted.
- sn: pd.DataFrame containing the conversion of stream coords to XY.
- a: string, name of the 'ds' coordinate column.
- b: string, name of the 'xs' coordinate column.
- X: string, name of the column with X coordinates.
- Y: string, name of the column with Y coordinates.

Returns:
- df: the dataframe with two extra columns X and Y
    
"""
def convertToXY(df, sn, a, b, X, Y):
    print(df.shape)
    sn = sn.rename(columns={'ds*': a, 'xs*': b, 'X': X, 'Y': Y})
    sn = sn.groupby([a, b], as_index=False).mean()
    df = pd.merge(df, sn[[a, b, X, Y]], on=[a, b], how='left') 
    print(df.shape)
    for i, row in df.iterrows():
        if row.isna().any():
            ds, xs = int(row[a]), int(row[b])
            df_p = sn.loc[(sn[a] == ds) & (abs(sn[b] - xs) == 1)]
            if not df_p.empty:
                df.loc[i, X] = df_p[X].mean()
                df.loc[i, Y] = df_p[Y].mean()
    return df

"""
Finds local maximums along the axis of dune crest direction (stream direction).

Parameters:
    bathies (ndarray): Array of bathymetry data.
    time (int): The index of the time step in the bathies array to be processed.
    sn (DataFrame): DataFrame containing the stream network data.
    order (int, optional): The order of the maximums to be identified. Default is 10.
    threshold_height (float, optional): The threshold height for local maximums to be considered. Default is 0.

Returns:
    crests (DataFrame): A DataFrame containing the local maximum points with XY coordinates, heights and cluster assignments.
"""
def find_local_maximums(bathies, time, sn, order=10, threshold_height=0):
    # Find local maximums along the axis of dune crest direction (stream direction)
    crest_indices = argrelextrema(bathies[time,4,:,:], np.greater, axis=1, order=order)
    
    # Create a DataFrame for all local maximums
    crests = {'ds':crest_indices[1], 'xs':crest_indices[0], 
              'ds*':crest_indices[1], 'xs*':crest_indices[0]+round(sn['xs'].min()),
              'Z':bathies[time,8,crest_indices[0], crest_indices[1]]}                   
    crests = pd.DataFrame(crests)
    
    # Add XY coordinates to crest points
    sn['ds*'], sn['xs*'] = round(sn['ds']), round(sn['xs'])
    crests = convertToXY(crests, sn, 'ds*', 'xs*' , 'X', 'Y')
    
    # Only take crests above a threshold value into account
    crests = crests[crests['Z'] >= threshold_height]
    
    # Group the crest points based on clustering
    crests = assign_dune_numbers(crests)
    
    # Remove non clustered outliers
    crests = crests.loc[crests['number'] != -1]
    crests = crests.reset_index(drop=True)
    
    return crests



"""
Reduces the number of crest points per dune by selecting only a subset of them based on a given interval.

Parameters:
crests (DataFrame): A dataframe containing the crest points to be reduced.
minimum_points: int, optional (default=3)
    The minimum amount of analysis points per dune.
interval: int, optional (default=20)
    The spacing between consecutive analysis points.
xs_min: float, optional (default=60)
    The minimum value of the 'xs' column to be selected.
xs_max: float, optional (default=200)
    The maximum value of the 'xs' column to be selected.
ds_min: float, optional (default=300)
    The minimum value of the 'ds' column to be selected.
ds_max: float, optional (default=2200)
    The maximum value of the 'ds' column to be selected.

Returns:
dunes (DataFrame): A dataframe containing the reduced set of crest points.
"""
def determine_analysis_points(crests, xs_min, xs_max, ds_min, ds_max, minimum_points=3, interval=20):
    dunes = pd.DataFrame()
    for i, dune in enumerate(crests['number'].unique()):
        df = crests.loc[crests['number'] == dune]
        # Make sure dunes with less than minimum_points crest points are removed
        if len(df) < minimum_points:
            continue
        # compute the step size, take at least minimum_points points for each dune
        step_size = min(interval, math.ceil(len(df)/(minimum_points)))
        # create a list of indices spaced out by the step size
        indices = list(range(0, df.shape[0], step_size))
        # select the rows from the dataframe using the indices
        selected_points = df.iloc[indices]
        # append the selected points to the new dataframe
        dunes = pd.concat([dunes, selected_points])
    # select only the rows where 'xs' and 'ds' are within the specified range
    dunes = dunes.loc[(dunes['xs'] >= xs_min) & (dunes['xs'] <= xs_max) & (dunes['ds'] >= ds_min) & (dunes['ds'] <= ds_max)]
    dunes = dunes.reset_index(drop=True)
    
    return dunes

"""
Generate a 4D array of bathymetry data from a collection of wavelet outputs.

Arguments:
file_info -- pandas.DataFrame with information about the input files
Znames -- list of names for the different versions of the bathymetry
Zavg_bed -- 2D numpy array of the average bed profile
Zmax_dict -- dictionary with maximum values of bathymetry for each time instance
Zmin_dict -- dictionary with minimum values of bathymetry for each time instance
wavelengths_all -- 1D numpy array of all wavelengths present in the wavelet outputs
lowerbound -- lower bound of the range of wavelengths to consider
upperbound -- upper bound of the range of wavelengths to consider
waveletOut_path -- path to the directory with the wavelet output files
hotspot -- name of the hotspot being analyzed

Returns:
4D numpy array with dimensions (time, Znames, xs, ds), containing the bathymetry data for each time instance.
"""
def generate_bathies(file_info, Znames, Zavg_bed, Zmax_dict, Zmin_dict, wavelengths_all, lowerbound, upperbound, waveletOut_path, hotspot):

    # Get the indices of values between the lower and upperbound wavelengths
    wavelengths = np.where((wavelengths_all >= lowerbound) & (wavelengths_all <= upperbound))[0]

    # Create a 4D array to store the data
    bathies = np.zeros((len(file_info), len(Znames), Zavg_bed.shape[0], Zavg_bed.shape[1]))

    # Loop over all time instances to generate a 4d matrix
    # [time, Znames, xs, ds]
    for i, time in enumerate(file_info['Date Time']):
        print(str(i) + '/' + str(len(file_info)))
        # Get the original different versions of the bathymetry 
        bathies[i,1,:,:] = Zmax_dict.get(time)                                  # Z with maximum values in overlapping cells [m+nap]
        bathies[i,2,:,:] = Zmin_dict.get(time)                                  # Z with minimum values in overlapping cells [m+nap]
        bathies[i,0,:,:] = (bathies[i,1,:,:] + bathies[i,2,:,:]) / 2            # Z with average values in overlapping cells [m+nap]
        bathies[i,3,:,:] = bathies[i,0,:,:] - Zavg_bed                          # Z minus the average bed profile [m]

        # Get wavelet output of this time instance
        rec3d = np.load(os.path.join(waveletOut_path,hotspot,str(i)+'.npy'))
        # Take only the wavelengths chosen between the lower and upper bound
        rec3d = np.take(rec3d, wavelengths, axis=2)
        # reconstruct the signal for t = i
        bathies[i,4,:,:] = reconstruct(rec3d, bathies[i,3,:,:])

        for j in range(4):
            bathies[i,j+5,:,:] = smoothen(bathies[i,j,:,:])
    
    return bathies

"""
Determine the length of dunes in the river axis direction based on the selected version of bathymetry.

Args:
    bathies (numpy.ndarray): 4D array of bathymetry data with dimensions (time, Znames, xs, ds).
    time (int): Index of the time instance to use for bathymetry data.
    dunes (pandas.DataFrame): DataFrame containing dune information with columns 'xs' and 'ds'.
    Zname (str, optional): Name of the version of bathymetry to use. Defaults to 'Zrec'.
    upperbound (float, optional): Upper bound of dune length in ds direction. Defaults to np.inf.
    lowerbound (float, optional): Lower bound of dune length in ds direction. Defaults to 0.
    
Returns:
    pandas.DataFrame: DataFrame with updated column 'length_ds' containing the determined length of dunes 
    in ds direction, and rows where length_ds is NaN or outside of the set boundaries removed.
"""
def get_dune_length_ds(bathies, time, dunes, upperbound, lowerbound, Zname='Zrec'):

    # Select version of the bathymetry to be used
    Znames = ['Z', 'Zmax', 'Zmin', 'Zavg', 'Zrec', 'Zf', 'Zmaxf', 'Zminf', 'Zavgf']
    Zname_index = Znames.index(Zname)

    length_ds = np.full((len(dunes),), np.nan)
    for index, row in dunes.iterrows():
        # Determine coordinates
        xs = int(row['xs'])
        ds = int(row['ds'])
        # Determine the coordinates of the troughs down- and upstream
        try:
            # Find all local minima at the line parallel to river axis through the current point
            minima = argrelextrema(bathies[time,Zname_index,xs,:], np.less, order=10)[0]
            # find ds value for through up and downstream
            min_up = minima[minima < ds].max()
            min_down = minima[minima > ds].min()
            # Determine dune length
            length_ds[index] = min_down - min_up
        except (ValueError, IndexError):
            # if this occurs, it means that there is not enough data to find both troughs
            continue

    # create new column in the dunes dataframe for the length in ds direction    
    dunes['length_ds'] = length_ds
    # drop rows where length_ds is NaN
    dunes.dropna(subset=['length_ds'], inplace=True)
    # drop rows where length_ds is outside of the set boundaries
    dunes = dunes.loc[(dunes['length_ds'] <= upperbound) & (dunes['length_ds'] >= lowerbound)]
    dunes = dunes.reset_index(drop=True)
    
    return dunes

"""
Rasterizes a list of polygons into a grid of the same shape using the polygon IDs as values.
The resulting raster is updated to restrict the number of overlapping polygons to at most 3.

Args:
polygons (GeoDataFrame): A GeoDataFrame containing the polygons to rasterize.
grid (ndarray): An array of the same shape as the desired output, representing the grid to rasterize the polygons onto.

Returns:
ndarray: A 2D numpy array of the same shape as the grid with the rasterized polygons.
"""
def rasterize_polygons(polygons, grid):
    # Define the transformation from coordinates to array indices
    transform = rasterio.Affine(1, 0, 0, 0, 1, 0)

    # Create a blank array of the same shape as the output grid
    count_array = np.zeros_like(grid, dtype=np.int16)

    # Loop over each polygon
    for geom, ID in zip(polygons.poly, polygons.index):
        if geom is None:
            continue
        # Rasterize the polygon into a temporary array
        temp_array = rasterio.features.rasterize(
            [(geom, 1)],
            out_shape=grid.shape,
            transform=transform,
            dtype=np.int16,
            fill=0,
            all_touched=True
        )

        # Add the temporary array to the count array
        count_array += temp_array

    # Replace negative values with 0
    count_array[count_array < 0] = 0

    return count_array

"""
Interpolates missing values in a 2D array by averaging neighboring values.

Args:
- arr (ndarray): The input 2D array with missing values.

Returns:
- ndarray: The array with missing values interpolated.

Algorithm:
- The function iterates over each element in the input array.
- If the element is NaN (missing value), it identifies the non-NaN neighboring values.
- If there are at least three valid neighbors, it replaces the missing value with their mean.
- The function returns the array with interpolated missing values.

Note:
- This function modifies the input array in place and also returns the modified array.
"""
def interpolate(arr):
    array = arr.copy()
    for i in range(array.shape[0]):
        for j in range(array.shape[1]):
            if np.isnan(array[i,j]):
                neighbors = []
                if i > 0 and not np.isnan(array[i-1,j]):
                    neighbors.append(array[i-1,j])
                if i < array.shape[0]-1 and not np.isnan(array[i+1,j]):
                    neighbors.append(array[i+1,j])
                if j > 0 and not np.isnan(array[i,j-1]):
                    neighbors.append(array[i,j-1])
                if j < array.shape[1]-1 and not np.isnan(array[i,j+1]):
                    neighbors.append(array[i,j+1])
                if len(neighbors) >= 3:
                    array[i,j] = np.mean(neighbors)
    return array

"""
Transform plough lines to stream coordinates and turn them into polygons by adding the plough width.

This function takes in plough lines represented as LineString geometries, extracts the x and y coordinates,
maps them to stream coordinates using a reference dataset, and converts them into polygons by adding a buffer
zone of half the plough width around the lines.

Args:
    lijnen (GeoDataFrame): The input GeoDataFrame containing the plough lines.
    sn (DataFrame): The reference dataset containing stream coordinates.
    plough_width (float): The plough width used to create buffer zones around the lines.
    xs_min (float): The minimum stream coordinate value.

Returns:
    GeoDataFrame: The output GeoDataFrame containing the transformed plough polygons.

"""
def plough_lines_to_polygons(lijnen, sn, plough_width, xs_min):
    
    # Create an empty list to store the transformed LineString geometries
    strco_line = [None] * len(lijnen)

    # Iterate over each plough line
    for i in range(len(lijnen)):
        arr = lijnen.iloc[i]['geometry']
        x_coords, y_coords = arr.xy

        ds_coords, xs_coords = np.full((len(x_coords),), np.nan), np.full((len(x_coords),), np.nan)
        j = 0
        for x, y in zip(x_coords, y_coords):
            try:
                sn_index = sn.loc[(sn['X'] == round(x)) & (sn['Y'] == round(y))].index[0]
                ds_coords[j] = sn.loc[sn_index, 'ds']
                xs_coords[j] = sn.loc[sn_index, 'xs']
                j += 1
            except IndexError:
                j += 1
                continue

        ds_coords = ds_coords[~np.isnan(ds_coords)]
        xs_coords = xs_coords[~np.isnan(xs_coords)]
        xs_coords -= xs_min
        
        if len(ds_coords) > 1 :
            strco_line[i] = LineString(list(zip(ds_coords, xs_coords)))

    # Create a new GeoDataFrame with the transformed LineString geometries and the date column
    plough_strco = gpd.GeoDataFrame({
        'geometry': strco_line,
        'date': lijnen['datum']
    })

    # Convert the date column to datetime format
    plough_strco['date'] = pd.to_datetime(plough_strco['date'], format='%Y%m%d')

    # Create a new column with polygons by buffering the LineString geometries
    plough_strco['poly'] = plough_strco['geometry'].buffer(plough_width/2)

    return plough_strco

"""
Calculate the quantity grid for dredging or dumping based on the given dictionary, file information, time, and bed elevation.

Args:
    dictionary (dict): The dictionary containing the data for dredging or dumping.
    file_info (DataFrame): The DataFrame containing file information.
    time (int): The time index for which to calculate the quantity grid.
    Zavg_bed (ndarray): The bed elevation grid.

Returns:
    ndarray: The quantity grid for dredging or dumping.

"""
def calculate_quantity_grid(dictionary, file_info, time, Zavg_bed):
    
    quantity_grid = np.full((Zavg_bed.shape[0], Zavg_bed.shape[1]), np.nan)
    
    if not time == 0:
        # Filter dictionary keys based on condition
        keys = [k for k in dictionary.keys() if file_info.loc[time-1, 'Date Time'] < k <= file_info.loc[time, 'Date Time']]
        if len(keys) > 0:
            print('leng keys long 0')
            df = pd.DataFrame()
            
            for key in keys:
                df = pd.concat([df, dictionary.get(key)])
                df.dropna(inplace=True)
                df['xs*'] = df['xs*'].astype(int)
                df['ds*'] = df['ds*'].astype(int)
            
            for i, row in df.iterrows():
                try:
                    quantity_grid[row['xs*']][row['ds*']] = row['volume']
                except IndexError:
                    continue
            
            quantity_grid = interpolate(quantity_grid)

    return quantity_grid

"""
Calculates the Normalized Spatial Cross Correlation on the fully bathymetry.

Performs the Normalized Spatial Cross Correlation on the fully bathymetry to determine the average
displacements between t = i and t = i+1.

Args:
    bathies (numpy.ndarray): The bathymetry data.
    time (int): The time value.
    time_diff_weeks (int): The time difference in weeks.

Returns:
    Tuple: A tuple containing the calculated values - (ccc, ds_shift_reg, ccf_reg)
"""
def nscc_full_bathymetry(bathies, time, time_diff_weeks):

    ds, xs = round(bathies.shape[3]/2), round(bathies.shape[2]/2)
    window_size_ds = ds
    window_size_xs = xs
    range_ds = max(50, int(100*time_diff_weeks))

    window0 = getWindow(bathies[time, 4, :, :], xs, ds, window_size_xs, window_size_ds)

    ccc = np.zeros(range_ds)
    xs1 = xs

    # Shifting
    for dds in range(range_ds):
        ds1 = ds + dds
        window1 = getWindow(bathies[time + 1, 4, :, :], xs1, ds1, window_size_xs, window_size_ds)

        # Report Rens van der Meijden - Normalized Cross Correlation
        top = (window0 - np.nanmean(window0)) * (window1 - np.nanmean(window1))
        bot1 = (window0 - np.nanmean(window0)) ** 2
        bot2 = (window1 - np.nanmean(window1)) ** 2
        ccc[dds] = np.nansum(top) / (np.nansum(bot1) * np.nansum(bot2)) ** 0.5

    ccc = np.nan_to_num(ccc, nan=0)
    ds_shift_reg = np.argmax(ccc)
    ccf_reg = ccc.max()

    return ds_shift_reg, ccf_reg

"""
Calculates the migration shift for each analysis point in the dunes dataframe.

Performs the migration shift calculation by using the Normalized Spatial Cross Correlation 
as described by Van der Meijden (2023), iterating over all analysis points represented by the rows of the
dunes dataframe. It determines the migration shift in the xs and ds directions, along with the correlation
coefficient for each point.

Args:
    bathies (numpy.ndarray): The bathymetry data.
    dunes (pandas.DataFrame): The dataframe containing the analysis points.
    time (int): The time value.
    ds_factor (int): The factor used for window size calculation based on dune length.
    range_xs (int): The search range in the xs direction.

Returns:
    pandas.DataFrame: The dunes dataframe with added columns for xs_shift, xs_new*, ds_shift, ds_new*, and ccf.
"""
def nssc_individual_analysis_points(bathies, dunes, time, ds_shift_reg, ds_factor, window_size_xs, Zname_index):

    # Determine search area (keep in mind, only xs is multiplied by 2)
    if ds_shift_reg < 20:
        # if mean shift is very small, set some default values for the search area
        range_ds_min, range_ds_max = 0, 40
        range_xs = 5
    else:
        # else set ds search area between 50% and 150% of the mean migration
        range_ds_min, range_ds_max = math.ceil(ds_shift_reg * 0.5), math.ceil(ds_shift_reg * 1.5)
        range_xs = math.ceil(ds_shift_reg/4)
    xs_shift = np.zeros(len(dunes))
    ds_shift = np.zeros(len(dunes))
    ccf = np.zeros(len(dunes))

    for i, point in dunes.iterrows():
        ds, xs = int(point['ds']), int(point['xs'])
        window_size_ds = math.ceil(point['length_ds'] / ds_factor)
        window0 = getWindow(bathies[time, Zname_index, :, :], xs, ds, window_size_xs, window_size_ds)

        ccc = np.zeros((range_xs * 2 + 1, range_ds_max - range_ds_min))

        for dds in range(range_ds_max - range_ds_min):
            ds1 = ds + range_ds_min + dds
            for dxs in range(range_xs * 2 + 1):
                xs1 = xs - range_xs + dxs

                window1 = getWindow(bathies[time + 1, Zname_index, :, :], xs1, ds1, window_size_xs, window_size_ds)
                # if window1.shape[1] - window0.shape[1] == 1:
                #     window1 = window1[:,:-1]
                # elif window0.shape[1] - window1.shape[1] == 1:
                #     window1 = np.append(window1, np.full((window1.shape[0], 1), np.nan), axis=1)
                try:    
                    top = (window0 - np.nanmean(window0)) * (window1 - np.nanmean(window1))
                    bot1 = (window0 - np.nanmean(window0)) ** 2
                    bot2 = (window1 - np.nanmean(window1)) ** 2
                    ccc[dxs, dds] = np.nansum(top) / (np.nansum(bot1) * np.nansum(bot2)) ** 0.5
                except ValueError:
                    continue

        ccc = np.nan_to_num(ccc, nan=0)
        xs_shift[i], ds_shift[i] = np.unravel_index(np.argmax(ccc), ccc.shape)
        ccf[i] = ccc.max()
        if i % 50 == 0:
            print(str(i) + '/' + str(len(dunes)))

    xs_shift -= range_xs
    ds_shift += range_ds_min
    dunes['xs_shift'] = xs_shift
    dunes['xs_new*'] = dunes['xs*'] + dunes['xs_shift']
    dunes['ds_shift'] = ds_shift
    dunes['ds_new*'] = dunes['ds*'] + dunes['ds_shift']
    dunes['ccf'] = ccf

    return dunes

def calculate_dune_parameters(dunes, bathies, dredging_grid, dumping_grid, plough_intensity, file_info, time, upperbound, lowerbound, threshold_height):
    """
    Calculates various parameters of dunes based on the given input data.
    
    Args:
        dunes (pandas.DataFrame): DataFrame containing dune data.
        bathies (numpy.ndarray): Array of bathymetry data.
        dredging_grid (numpy.ndarray): Array of dredging data.
        dumping_grid (numpy.ndarray): Array of dumping data.
        plough_intensity (numpy.ndarray): Array of plough intensity data.
        time (int): Index of the current time in the data.
        upperbound (float): Upper bound value for coordinate calculations.
        lowerbound (float): Lower bound value for coordinate calculations.
        threshold_height (float): Threshold height for determining local maxima and minima.
    
    Returns:
        pandas.DataFrame: DataFrame with calculated dune parameters added as columns.
    """
    # Initiate empty arrays for determining dune parameters
    length = np.full((len(dunes),), np.nan)
    height = np.full((len(dunes),), np.nan)
    lee_slope = np.full((len(dunes),), np.nan)
    dredged_volume = np.full((len(dunes),), np.nan)
    dredged_height = np.zeros(len(dunes))
    dumped_volume = np.full((len(dunes),), np.nan)
    ploughed_sum = np.full((len(dunes),), 0)
    ploughed_max = np.full((len(dunes),), 0)
    heightNAP = np.full((len(dunes),), np.nan)
    dune_volume = np.full((len(dunes),), np.nan) 
    
    for point_index in range(len(dunes)):
        
        # if not point_index == 34:
        #     continue
        
        # Define the two points you want to connect
        point1 = (dunes.loc[point_index, 'ds'], dunes.loc[point_index, 'xs'])
        point2 = (dunes.loc[point_index, 'ds']+dunes.loc[point_index, 'ds_shift'], dunes.loc[point_index, 'xs']+dunes.loc[point_index, 'xs_shift'])
        
        if np.isnan(bathies[time, 4, int(point1[1]), int(point1[0])]):
            print('analysis point outside of bathymetry')
            continue
        
        # Calculate the distance and angle between the two points
        dds = point2[0] - point1[0]
        dxs = point2[1] - point1[1]
        distance = np.sqrt(dds**2 + dxs**2)
        angle = np.arctan2(dxs, dds)
        
        # Calculate the ds and xs coordinates of each point along the line
        ds_coords = np.linspace(point1[0] - upperbound/2*np.cos(angle), point1[0] + upperbound/2*np.cos(angle), num=301, endpoint=True)
        xs_coords = np.linspace(point1[1] - upperbound/2*np.sin(angle), point1[1] + upperbound/2*np.sin(angle), num=301, endpoint=True)
        
        try:
            # Reconstructed wavelet signal along the line
            elevation_rec = [bathies[time, 4, int(round(xs)), int(round(ds))] for ds, xs in zip(ds_coords, xs_coords)]
            elevation_rec = np.array(elevation_rec)
            elevation_rec = savgol_filter(elevation_rec,7,3,mode='mirror',axis = 0)
            
            elevation_orig = [bathies[time, 3, int(round(xs)), int(round(ds))] for ds, xs in zip(ds_coords, xs_coords)]
            elevation_orig = np.array(elevation_orig)
            elevation_orig = savgol_filter(elevation_orig,7,3,mode='mirror',axis = 0)
            
            dredging = [dredging_grid[int(round(xs)), int(round(ds))] for ds, xs in zip(ds_coords, xs_coords)]
            dredging = np.array(dredging)
            
            dumping = [dumping_grid[int(round(xs)), int(round(ds))] for ds, xs in zip(ds_coords, xs_coords)]
            dumping = np.array(dumping)
            
            ploughed = [plough_intensity[int(round(xs)), int(round(ds))] for ds, xs in zip(ds_coords, xs_coords)]
            ploughed = np.array(ploughed)
            
            nonaveraged = [bathies[time, 5, int(round(xs)), int(round(ds))] for ds, xs in zip(ds_coords, xs_coords)]
            nonaveraged = np.array(nonaveraged)
            
        except IndexError:
            # This error is fine. This means that spatial cross correlation result is out of bounds.
            # Just skip this data.
            print('IndexError')
            continue
        
        # Determine the coordinates of the crest and troughs on reconstructed profile
        maxima = argrelextrema(elevation_rec, np.greater, order=10)[0]
        maxima = maxima[elevation_rec[maxima] > threshold_height]
        
        # Determine the coordinates of the troughs down- and upstream
        minima = argrelextrema(elevation_rec, np.less, order=10)[0]
        minima = minima[elevation_rec[minima] < -threshold_height]
        try:
            max_xr = maxima[np.abs(maxima - upperbound/2).argmin()]

            min_up_xr = minima[minima < upperbound/2].max()
            min_down_xr = minima[minima > upperbound/2].min()
        except ValueError:
            print('No local min or max found on at least one side of the analysis point')
            continue
        
        max_yr = elevation_rec[max_xr]
        
        min_up_yr = elevation_rec[min_up_xr]
        min_down_yr = elevation_rec[min_down_xr]
        
        # Check if crest point is close to the analysis point
        if abs(max_xr - upperbound/2) > 50:                                        #!!! Check this value
            print('closest crest too far from analysis point. PI:_' + str(point_index))
            continue
        
        # Determine the coordinates of the crest and troughs on the original profile
        max_xo = np.argmax(elevation_orig[min_up_xr:min_down_xr]) + min_up_xr
        max_yo = elevation_orig[max_xo]
        
        try:
            min_up_xo = np.nanargmin(elevation_orig[int(min_up_xr-lowerbound/2):int(min_up_xr+lowerbound/2)]) + int(min_up_xr-lowerbound/2)
            min_down_xo = np.nanargmin(elevation_orig[int(min_down_xr-lowerbound/2):int(min_down_xr+lowerbound/2)]) + int(min_down_xr-lowerbound/2)
        except ValueError:
            continue
        min_up_yo = elevation_orig[min_up_xo]
        min_down_yo = elevation_orig[min_down_xo]
        
        # Check if there is no elevation data missing between the troughs
        if np.isnan(elevation_orig[min_up_xr:min_down_xr]).any():
            print('elevation profile between troughs is discontinuous')
            continue
        
        # Determine dune length and height
        length[point_index] = min_down_xo - min_up_xo
        height[point_index] = max_yo - (min_down_yo + min_up_yo)/2
        
        # Determine lee angle
        # defined as the mean slope of the middle 4/6th part of the dune lee slope (van der Mark et al., 2008)
        # echte horizontaal ivm zwaartekracht
        lee1_x, lee2_x = round(max_xo + (min_down_xo - max_xo)/6), round(max_xo + (min_down_xo - max_xo)*5/6)
        lee1_y, lee2_y = elevation_orig[lee1_x], elevation_orig[lee2_x]
        lee_slope[point_index] = np.arctan2((lee2_y-lee1_y),(lee2_x-lee1_x))
        
        # Determine dredged and dumped volume [in m3/m]
        dredged_volume[point_index] = np.nansum(dredging[min_up_xo:min_down_xo])
        try:
            H_dredged = np.nanmax(dredging[min_up_xo:min_down_xo])
        except ValueError:
            H_dredged = np.nan
        if not np.isnan(H_dredged):
            dredged_height[point_index] = H_dredged
        dumped_volume[point_index] = np.nansum(dumping[min_up_xo:min_down_xo])
        
        # Determine plough quantity
        ploughed_sum[point_index] = np.nansum(ploughed[min_up_xo:min_down_xo])
        try:
            max_plough = np.nanmax(ploughed[min_up_xo:min_down_xo])
        except ValueError:
            max_plough = np.nan
        if not np.isnan(max_plough): 
            ploughed_max[point_index] = max_plough
        
        # Determine highest point w.r.t NAP
        try:
            heightNAP[point_index] = np.nanmax(nonaveraged[min_up_xo:min_down_xo])
        except:
            heightNAP[point_index] = np.nan
        # Determine dune volume
        floor = np.linspace(min_up_yo,min_down_yo, min_down_xo-min_up_xo+1)
        dune_volume[point_index] = (elevation_orig[min_up_xo:min_down_xo+1] - floor).sum()

    
    dunes['length'] = length
    dunes['height'] = height
    dunes['aspect ratio'] = height / length
    dunes['lee angle'] = np.degrees(lee_slope)
    
    dunes['Zmax_NAP'] = heightNAP
    dunes['dune volume'] = dune_volume
    
    dunes['dredged volume'] = dredged_volume
    dunes['dredged height'] = dredged_height
    dunes['dumped volume'] = dumped_volume
    dunes['ploughed_sum'] = ploughed_sum
    dunes['ploughed_max'] = ploughed_max
    
    dunes['time index'] = time
    dunes['time'] = file_info.loc[time, 'Date Time']
    
    return dunes

















