# -*- coding: utf-8 -*-
"""
Created on Mon Mar 13 14:29:48 2023

@author: kruis
"""

#%% Initiate script, imports and paths
import os
import pandas as pd
import numpy as np
import json

def detect_delimiter(file_path):
    """
    Detects the delimiter sign of a raw data file.
    
    Args:
        file_path (String): Path to the directory with the filename.
        
    Returns:
        String: Delimiter sign.
    """
    with open(file_path, 'r') as f:
        line = f.readline()
        if ',' in line:
            return ','
        else:
            return ' '

# Path to the directory with raw data files
folder_path = "D:/_metingen_DTON" 
# Path to the folder where the processed MBES data should be saved in JSON files
outFolder = r'C:/Users/kruis/Documents/Thesis/Python/processedMBES'

#%% Load important data

# Loop through all files in the folder and add their paths to the list
file_list = []
for filename in os.listdir(folder_path):
    file_path = os.path.join(folder_path, filename)
    if os.path.isfile(file_path) and "Hotspots_" in filename and len(filename) >= 12 and filename[:6].isdigit() and filename[7:11].isdigit(): # check if the filename has a valid datetime format and contains "Hotspots_"
        hotspot_letter = filename.split("Hotspots_")[1][0] # extract the character after "Hotspots_"
        tp_value = 1 if "TP" in filename else 0 # check if "TP" is in the filename and assign 1 or 0 accordingly
        file_list.append((filename[:], pd.to_datetime(filename[:11], format='%y%m%d_%H%M'), hotspot_letter, tp_value))

# Create a data frame from the file list
df = pd.DataFrame(file_list, columns=['File Name', 'Date Time', 'Hotspot Letter', 'TP Value'])
# Filter dataframe for files that are not specific to a hotspot
df = df[df['File Name'].str.contains('ABC') == False]
df = df[df['Hotspot Letter'].str.match('[A-J]')]

# Store the dataframe with all info about the seperate raw mbes files in a .pkl format
df.to_pickle(os.path.join(outFolder,'file_info.pkl'))

# Get a dataframe that contains hotspot specific information, such as X, Y locations
hotspot_data = pd.read_excel(os.path.join(folder_path, 'hotspot_data.xlsx'))

# Define the size of the grid cells 
cell_size = 1.0

#%% Processing

# Loops over the 10 hotspots [A-J], index format
for hs in range(len(hotspot_data)):
    
    # Indicator for the current hotspot that is being processed
    print(hotspot_data.loc[hs, 'Hotspot'])
    
    # Create dataframe with just the measurements from hotspot the current hotspot   
    df_HS = df[df['Hotspot Letter'] == hotspot_data.loc[hs, 'Hotspot']]
    df_HS.reset_index(inplace=True, drop=True)
    
    # Determine the minimum and maximum x and y values in the hotspot from QGis
    min_x, max_x = hotspot_data.loc[hs, 'xmin'], hotspot_data.loc[hs, 'xmax']
    min_y, max_y = hotspot_data.loc[hs, 'ymin'], hotspot_data.loc[hs, 'ymax']
    
    # Compute the number of rows and columns in the grid
    num_rows = int(np.ceil((max_y - min_y) / cell_size))
    num_cols = int(np.ceil((max_x - min_x) / cell_size))
    
    # Create an empty dictionary to store all processed MBES data files of the hotspot
    hotspot_dict = {}
    
    # Loops over all raw data files that are specific to the current hotspot
    for i in range(len(df_HS)):
        
        # Progress indicator
        print(round(i/len(df_HS)*100)+'%')
        
        # Initialize an array to store the average z values for each cell
        z_values = np.zeros((num_rows, num_cols))
        
        # Set a variable which can be used to skip a file if it doesnot store the data in a processing friendly way
        skip = 0
        
        # Load raw MBES data file as dataframe
        file = df_HS.loc[i, 'File Name']
        delimiter = detect_delimiter(os.path.join(folder_path, file))
        data = pd.read_csv(os.path.join(folder_path, file), sep=delimiter, header=None, names=['x', 'y', 'z'])
        
        # Loops over all rows of the file. In this case, over all X, Y, Z coordinates
        # Several exceptions are thrown such that the script does not break due to a corrupt file
        # Currently, approximately 2% of files is skipped
        for row in range(len(data)):
            try:
                x = int(data.loc[row, 'x'])
                y = int(data.loc[row, 'y'])
                z = data.loc[row, 'z']
            except KeyError:
                print("A Key error was detected")
                skip = 1
                break
            except ValueError:
                print("A Value error was detected")
                skip = 1
                break
            else:    
                try:
                    z_values[y-min_y, x-min_x] = z
                except IndexError:
                    print("Index out of bounds")
                    skip = 1
                    break 
        
        # If file is fully processed, the resulting 2D np.Array is storred in the dictionary
        # with the datetime as key
        if skip == 0:
            hotspot_dict.update({df_HS.loc[i, 'Date Time']: z_values})
        
    # Store the hotspot dictionary to a .JSON file. In this way, the file can be accessed
    # easily after processing. Datetime objects and arrays are first serialized to string 
    # and list formats. These must be deserialized when loaded back.
    # Convert datetime keys and numpy arrays to serializable format
    serialized_dict = {}
    for key, value in hotspot_dict.items():
        serialized_dict[str(key)] = value.tolist()

    # Open a file for writing
    with open(os.path.join(outFolder,hotspot_data.loc[hs, 'OutFolder']), 'w') as f:
        # Write the dictionary to the file in JSON format
        json.dump(serialized_dict, f)

